#include <os.h>
#include <SDL/SDL.h>
#include <SDL/SDL_gfxPrimitives.h>
#include "Tetromino.h"
#include <libndls.h>

SDL_Surface *screen;
nSDL_Font *font;
nSDL_Font *font2;
SDL_bool done = SDL_FALSE;

#define TETRIS_VERSION "1.0a"

#define BOARD_HEIGHT 21       //20 rows + 1 row at the very bottom (lower wall)
#define BOARD_WIDTH 12        //10 columns + 1 column on the left (left wall) + 1 column on the right (right wall)


int tetroX = BOARD_WIDTH/2;
int tetroY = 0;
int tetroR = 0;
int tetroT = 1;

int tetrisBoard[BOARD_WIDTH][BOARD_HEIGHT];

int decalXboard = 50;
int decalYboard = 10;
int cellsize = 10;

Uint32 timerstart=0;
Uint32 currenttimer=0;

int level = 1;
int linesremovedtotal = 0;
int linesremovedlevel = 0;
int linesremoved = 0;
int score = 0;


void init_board( void )
{
    // Fill the Board with correct values :
    // -2 when we are in presence of a wall
    // -1 when the board call is empty
    for( int i=0; i<BOARD_WIDTH; i++)
    {
        for( int j=0; j<BOARD_HEIGHT; j++)
        {
            tetrisBoard[i][j] = -1;             // We are in the middle of the board -> empty cell
        }      
    }   
    
    for( int j=0; j<BOARD_HEIGHT; j++)
    {
        tetrisBoard[0][j] = -2;                 // We are on the very left -> wall
        tetrisBoard[BOARD_WIDTH-1][j] = -2;     // We are on the very right -> wall
    }

    for( int i=0; i<BOARD_WIDTH; i++)
    {
        tetrisBoard[i][BOARD_HEIGHT-1] = -2;    // We are at the bottom row -> wall      
    }      
}

void fill_board_random( void )
{
    // Fill the Board with correct values :
    // -2 when we are in presence of a wall
    // -1 when the board call is empty
    for( int i=0; i<BOARD_WIDTH; i++)
    {
        for( int j=0; j<BOARD_HEIGHT; j++)
        {
            tetrisBoard[i][j] = rand()%7;             // We are in the middle of the board -> we fill with random color
        }      
    }   
    
    for( int j=0; j<BOARD_HEIGHT; j++)
    {
        tetrisBoard[0][j] = -2;                 // We are on the very left -> wall
        tetrisBoard[BOARD_WIDTH-1][j] = -2;     // We are on the very right -> wall
    }

    for( int i=0; i<BOARD_WIDTH; i++)
    {
        tetrisBoard[i][BOARD_HEIGHT-1] = -2;    // We are at the bottom row -> wall      
    }      
    
    #ifdef DEBUG
        FILE * pFile;
        pFile = fopen ("/documents/ProgSly/Board.tns","w");
        if (pFile!=NULL)
        {
            for( int i=0; i<BOARD_HEIGHT; i++)
            {
                for( int j=0; j<BOARD_WIDTH; j++)
                {
                    fprintf(pFile, "%d ", tetrisBoard[j][i]);
                }
                fprintf(pFile, "\n"); 
            }
            fclose (pFile);
        }
    #endif

}

void init_SDL(void)
{
    if(SDL_Init(SDL_INIT_VIDEO) == -1)
    {
        printf("Sorry, can't init SDL: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
    screen = SDL_SetVideoMode(320, 240, has_colors ? 16 : 8, SDL_SWSURFACE);
    if(screen == NULL)
    {
        printf("Sorry, can't init display: %s\n", SDL_GetError());
        SDL_Quit();
        exit(EXIT_FAILURE);
    }
    SDL_ShowCursor(SDL_DISABLE);
    
    font = nSDL_LoadFont(NSDL_FONT_TINYTYPE, 255, 255, 255);
    font2 = nSDL_LoadFont(NSDL_FONT_TINYTYPE, 255, 0, 0);
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
}

void quit(void)
{
    SDL_Quit();
}

void draw_score_zone( void )
{
    int widthtext=0;
    char buffer[15];
    
    boxRGBA( screen, 200, 20, 300, 50, 128, 128, 128, 255 );
    rectangleRGBA( screen, 200, 20, 300, 50, 255, 255, 255, 255 );
    nSDL_DrawString(screen, font2, 210, 25, "Simple TETRIS " );
    nSDL_DrawString(screen, font, 210, 35, "Version  " TETRIS_VERSION );
        
    boxRGBA( screen, 200, 70, 300, 170, 128, 128, 128, 255 );
    rectangleRGBA( screen, 200, 70, 300, 170, 255, 255, 255, 255 );
    nSDL_DrawString(screen, font2, 210, 75, "Score");
    nSDL_DrawString(screen, font2, 210, 105, "Level");
    nSDL_DrawString(screen, font2, 210, 135, "Lines");   
    
    sprintf( buffer, "%d", score);
    widthtext=nSDL_GetStringWidth(font, buffer );
    nSDL_DrawString(screen, font, 290-widthtext, 90, buffer );  
    
    sprintf( buffer, "%d", level);
    widthtext=nSDL_GetStringWidth(font, buffer );
    nSDL_DrawString(screen, font, 290-widthtext, 120, buffer );  
    
    sprintf( buffer, "%d", linesremovedtotal);
    widthtext=nSDL_GetStringWidth(font, buffer );
    nSDL_DrawString(screen, font, 290-widthtext, 150, buffer );  

}

void draw_board(void)
{
    Uint8 R,G,B,A;
    
    for( int i=0; i<BOARD_WIDTH; i++)
    {
        for( int j=0; j<BOARD_HEIGHT; j++)
        {
            // We get the content of the cell
            int ghosttetrotype = tetrisBoard[i][j];
            
            if (ghosttetrotype==-2)
            {
                //And we pick the corresponding color : !! Be carefull, the first 3 colors are for the Grid/walls/board background
                R=TetroColors[1][0];
                G=TetroColors[1][1];
                B=TetroColors[1][2];
                A=TetroColors[1][3];
            }
            else if (ghosttetrotype==-1)
            {
                //And we pick the corresponding color : !! Be carefull, the first 3 colors are for the Grid/walls/board background
                R=TetroColors[2][0];
                G=TetroColors[2][1];
                B=TetroColors[2][2];
                A=TetroColors[2][3];
            }
            else
            {
                //And we pick the corresponding color : !! Be carefull, the first 3 colors are for the Grid/walls/board background
                R=TetroColors[ghosttetrotype+3][0];
                G=TetroColors[ghosttetrotype+3][1];
                B=TetroColors[ghosttetrotype+3][2];
                A=TetroColors[ghosttetrotype+3][3];
            }            
            //and we draw the cell with the according color
            boxRGBA( screen, i*cellsize+decalXboard, j*cellsize+decalYboard, (i+1)*cellsize+decalXboard, (j+1)*cellsize+decalYboard, R, G, B, A );
            
            //And we pick the color of the grid (the first one in the list)
            R=TetroColors[0][0];
            G=TetroColors[0][1];
            B=TetroColors[0][2];
            A=TetroColors[0][3];
        
            //and we draw the cell with the according color
            rectangleRGBA( screen, i*cellsize+decalXboard, j*cellsize+decalYboard, (i+1)*cellsize+decalXboard, (j+1)*cellsize+decalYboard, R, G, B, A );
        }
    } 
}

void draw_tetromino( void )
{
    //Pick the color parameters for that Tetromino
    Uint8 R=TetroColors[tetroT+3][0];
    Uint8 G=TetroColors[tetroT+3][1];
    Uint8 B=TetroColors[tetroT+3][2];
    Uint8 A=TetroColors[tetroT+3][3];
    
    //Pick the line color (the first one)
    Uint8 Rl=TetroColors[0][0];
    Uint8 Gl=TetroColors[0][1];
    Uint8 Bl=TetroColors[0][2];
    Uint8 Al=TetroColors[0][3];
    
    for(int j=0; j<4; j++)
    {
        for(int i=0; i<4; i++)
        {
            if (Tetrominos[tetroT][tetroR][i+j*4] == 1)
            {
                boxRGBA( screen, (tetroX+i)*cellsize+decalXboard, (tetroY+j)*cellsize+decalYboard, (tetroX+i+1)*cellsize+decalXboard, (tetroY+j+1)*cellsize+decalYboard, R, G, B, A );
                rectangleRGBA( screen, (tetroX+i)*cellsize+decalXboard, (tetroY+j)*cellsize+decalYboard, (tetroX+i+1)*cellsize+decalXboard, (tetroY+j+1)*cellsize+decalYboard, Rl, Gl, Bl, Al );
            }         
        }      
    }
}

bool is_line_full( int line )
{
    for(int i=1; i<BOARD_WIDTH-1; i++)
    {
        if (tetrisBoard[i][line]==-1) return false; //There is at least one free cell in this line
    }
    return true; //All cells are filled so line is full --> we can remove it and score !!
}

void remove_line_from_board( int line )
{
    if (line==0)
    {
        for(int i=1; i<BOARD_WIDTH-1; i++)
        {
            tetrisBoard[i][0]=-1;   //If the top line is full, just remove all the values by setting -1 instead (i.e blank cell)
        }
    }
    else
    {
        for(int j=line; j>0; j--)
        {
            for(int i=1; i<BOARD_WIDTH-1; i++)
            {
                tetrisBoard[i][j]=tetrisBoard[i][j-1];   //We go from down to up and shift the cells one row below
            }
        }
        for(int i=1; i<BOARD_WIDTH-1; i++)
        {
            tetrisBoard[i][0]=-1;   // and finaly we create a blank line at the top of the board
        }       
    }
}

int do_we_have_full_line( void )
{
    int nblinetoremove = 0;

    for(int i=0; i<BOARD_HEIGHT-1; i++)
    {
        if (is_line_full(i))
        {
            nblinetoremove++;   //will be used for the score when several lines are filled at the same timerstart
            remove_line_from_board(i);
        }
    }
    return nblinetoremove;
}

void compute_score( int line )
{
        switch(line)
        {
            case 0: 
                break;
            
            case 1: 
                score+=1*level; 
                linesremovedtotal+=1; 
                linesremovedlevel+=1;
                break;
            
            case 2: 
                score+=2*level; 
                linesremovedtotal+=2;
                linesremovedlevel+=2;
                break;
                
            case 3: 
                score+=5*level; 
                linesremovedtotal+=3;
                linesremovedlevel+=3;
                break;
                
            case 4: 
                score+=10*level; 
                linesremovedtotal+=4;
                linesremovedlevel+=4;
                break;
                
            default: 
                break;
        }
        
        if (linesremovedlevel>=10)
        {
            level++;
            linesremovedlevel=0;
        }
}

void we_reach_bottom( void )
{
    //this is the routine that is called when the tetromino cannot go lower
    //so we transfer it to the board
    for(int j=0; j<4; j++)
    {
        for(int i=0; i<4; i++)
        {
            if (Tetrominos[tetroT][tetroR][i+j*4] == 1)
            {
                tetrisBoard[tetroX+i][tetroY+j]=tetroT;    // We fix the tetromino type in the board
            }         
        }      
    }
}

bool do_fit( int X, int Y, int T, int R )
{
    for(int j=0; j<4; j++)
    {
        for(int i =0; i<4; i++)
        {
            if ((Tetrominos[T][R][i+j*4]==1) && (tetrisBoard[X+i][Y+j]!=-1)) return false;
        }
    }
    return true;
}

void you_loose()
{
    fill_board_random();
    draw_board();
    
    boxRGBA( screen, 40, 80, 300, 160, 0, 0, 0, 255 );
    rectangleRGBA( screen, 10, 10, 310, 230, 0, 0, 0, 255 );   
    
    nSDL_DrawString(screen, font2, 50, 85, "GAME OVER !!!");
    nSDL_DrawString(screen, font, 50, 100, "Your score : %d", score);
    nSDL_DrawString(screen, font, 50, 115, "Lines : %d", linesremovedtotal);
    nSDL_DrawString(screen, font, 50, 130, "Level : %d", level);
    nSDL_DrawString(screen, font2, 50, 145, "PRESS A KEY TO QUIT");
    
    SDL_Flip(screen);
            
    wait_key_pressed();
    done=SDL_TRUE;
}

void handle_keydown(SDLKey key)
{
    switch(key)
    {
        case SDLK_TAB:      //Ask for a rotation iof the tetromino
                if (do_fit( tetroX, tetroY, tetroT, (tetroR+1)%4 )==true) tetroR = (tetroR+1)%4;
            break;
               
        case SDLK_DOWN:
                if (do_fit( tetroX, tetroY+1, tetroT, tetroR )==true) tetroY++;
            break;
        
        case SDLK_LEFT:
                if (do_fit( tetroX-1, tetroY, tetroT, tetroR )==true) tetroX--;
            break;
        
        case SDLK_RIGHT:
                if (do_fit( tetroX+1, tetroY, tetroT, tetroR )==true) tetroX++;
            break;
               
        case SDLK_ESCAPE:
            done = SDL_TRUE;
            break;
        
        default:
            break;
    }
}


int main(void)
{
    int Dt[]={ 1000, 900, 800, 750, 700, 650, 600, 550, 500, 450, 400, 350, 300, 250, 200 };
    int deltatimefall;
    
    SDL_Event event;
    
    init_SDL();  
    init_board();
    draw_board();
    
    timerstart = 0;
    
    srand( timerstart );
    
    while(!done)
    {
        draw_board();
        draw_tetromino();
        draw_score_zone();
        
        SDL_Flip(screen);
        
        while(SDL_PollEvent(&event))
        {
            switch(event.type)
            {
                case SDL_KEYDOWN:
                    handle_keydown(event.key.keysym.sym);
                    break;
                    
                default:
                    break;
            }
        }
        
        currenttimer=SDL_GetTicks();
        if (level<=15) deltatimefall = Dt[level-1];
        else deltatimefall = Dt[14];
                
        if( currenttimer > (timerstart+deltatimefall) )
        {
            if (do_fit( tetroX, tetroY+1, tetroT, tetroR )==true)
            {
                tetroY++;
            }
            else
            {
                we_reach_bottom(); // we cannot go below, so we need to stop the tetromino
                
                // and we launch a new tetromino
                tetroX = 4;
                tetroY = 0;
                tetroR = 0;
                tetroT = rand()%7;
                
                if(do_fit( tetroX, tetroY, tetroT, tetroR )!=true)
                {
                    you_loose();
                }
                
            }
            
            timerstart = currenttimer;  //we restart timer
        }
        
        linesremoved = do_we_have_full_line();
        
        compute_score( linesremoved );
        
    }
    
    nSDL_FreeFont(font);
    nSDL_FreeFont(font2);
    
    quit();
    
    return EXIT_SUCCESS;
}
